/* amphibian routines.  some of these should probably end up
   implemented by students instead of in this file. */ 
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/un.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h> /* NOFILE */
#include <assert.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>

#include "fishguts.h"
#include "amphibian.h"

#define min(x,y) ({ \
 const typeof(x) _xi = (x);       \
 const typeof(y) _yi = (y);       \
 (void) (&_xi == &_yi);           \
 _xi < _yi ? _xi : _yi; })
#define max(x,y) ({ \
 const typeof(x) _xa = (x);       \
 const typeof(y) _ya = (y);       \
 (void) (&_xa == &_ya);           \
 _xa > _ya ? _xa : _ya; })

/* The protocol structure consisting of function pointers
   that implement connect, close, write, and bind. */
struct amphibian_protocol *proto;

struct transport_state; 

/* the type of a signal handler function */
typedef void (*siggy_ztardust)(int);

/* we have to know the transport_app_state structure for a
   socket (the reverse mapping is stored inside the
   transport state's app structure) */
static struct transport_app_state *ts_array[NOFILE];
static void 
set_transport_app_state_for(int connected_socket, 
                        struct transport_app_state *app) {
  ts_array[connected_socket] = app;
}
static struct transport_app_state * 
get_transport_app_state_for(int connected_socket) {
  return ts_array[connected_socket];
}

/* amph_close makes sure all the state has gone away */
void amph_close(int fd);

/* when the application calls 'write' this is what we do */
static void amphibian_app_write(int connected_sock);

/* called when we're ready to receive new data from the
   application.  this happens when we receive an ack from
   the other side. */
static void 
enable_readhook(struct transport_app_state *app) {
  /* sometimes we have to stop paying attention to the socket, as we're blocked
     on retransmissions.  this undoes that on receiving an ack */
  fish_readhook(app->fd, amphibian_app_write);
  app->ready_for_data=NULL;
}

/* called when we receive ack-of-syn.  Tell the app that all
   is good. */
static void 
amph_connection_established(struct transport_app_state *app) {
  /* don't return from fishsocket_connect until we get an
     ack from the other side. */
  int len = strlen(CONNECT_SUCCESS);
  fprintf(stderr, "%s\n", __FUNCTION__);
  if(write(app->fd, CONNECT_SUCCESS, len) < len) {
    amph_close(app->fd);
    fprintf(stderr, "%s closing\n", __FUNCTION__);
    return;
  }
  enable_readhook(app);
}

/* the application has called write.  we now grab the data, and 
   try to send it across the network.  NOTE that if we're already
   sending data, we'll not read new data from the socket.  The 
   transport protocol is stop-and-wait, so it would get discarded.
*/
static void amphibian_app_write(int connected_sock) {
  struct transport_app_state *app = get_transport_app_state_for(connected_sock);
  int can_be_written = proto->avail_write_buffer(app); 
  assert(proto != NULL);
  
  /* if the transport protocol has send buffer space available */
  if(can_be_written > 0) {
    /* read into the buffer */
    int buffer_size = min(can_be_written, 32 * 1024);
    char *buf = malloc(buffer_size);
    int readbytes = read(connected_sock, buf, buffer_size);
    if(readbytes == 0) {
      /* the application has called close() or shutdown(SHUT_WR) */
      proto->close( app );
      fish_remove_readhook(connected_sock);
    } else if(readbytes > 0) {
      /* there was some data the application sent */
      proto->write( app, buf, readbytes);
    } else {
      fprintf(stderr, "error on socket %d: %s; dropping", connected_sock,
              strerror(errno));
      fish_remove_readhook(connected_sock);
      close(connected_sock);
    }
    free(buf);
  } else {
    fish_remove_readhook(connected_sock);
    app->ready_for_data=enable_readhook;
  }
}

/* this is an esoteric and strange function I hope no one is
   ever expected to understand.  The point is that it's
   possible to send file descriptors over unix sockets. */
/* from http://www.wcug.wwu.edu/lists/netdev/199802/msg00143.html Andi Kleen */
/* with non-posix code from http://blade.nagaokaut.ac.jp/ruby/ruby-dev/16068 */
static int sfd(int passfd, int fd, int data)
{
  char cbuf[CMSG_SPACE(sizeof(int))]; 
  struct msghdr mh = {0,0,0,0,0,0,0};
  struct cmsghdr *cm;
  int *dp; 
  struct iovec iov;

  assert(fd != -1);

  if (fd >= 0) { 
#if defined(HAVE_ST_MSG_CONTROL) || defined(__linux__)
    mh.msg_control = cbuf; // (caddr_t)&cmsg;
    mh.msg_controllen = sizeof cbuf; // sizeof(struct cmsghdr) + sizeof(int);
    mh.msg_flags = 0;
    cm = CMSG_FIRSTHDR(&mh); 
    cm->cmsg_len = sizeof(struct cmsghdr) + sizeof(int);
    cm->cmsg_level = SOL_SOCKET;
    cm->cmsg_type = SCM_RIGHTS;
    dp = (int *)(cm+1); /* CMSG_DATA(cm) is correct, but generates an alignment warning  */
    *dp = fd;
#else
    mh.msg_accrights = (caddr_t)&fd;
    mh.msg_accrightslen = sizeof(fd);
#endif
  }
  if (data != 0) {
    iov.iov_base = &data; 
    iov.iov_len = sizeof(int); 
    mh.msg_iov = &iov; 
    mh.msg_iovlen = 1; 
  }

  return sendmsg(passfd, &mh, 0);
}

/* an attempt at a closer; not necessarily any good */
void amph_close(int fishsocket) {
  struct transport_app_state *app = get_transport_app_state_for(fishsocket);
  assert(proto != NULL);
  if(app!=NULL) {
    proto->close(app);
  }
  fish_remove_readhook(fishsocket);
  set_transport_app_state_for(fishsocket, NULL);
}

/* we're about to accept a new socket on behalf of the
   application, so we have to pass it a file descriptor.  we
   create an anonymous socket, send it, and then register
   our interest in data the app writes */
int amphibian_accept_for_app(int connected_sock) {
  /* this speech is my recital, I think it's very vital
     to rock a rhyme, that's right on time.
     it's tricky is the title, here we go...  - run dmc */
  int socket_vector[2];
  int zok;
  siggy_ztardust old_sigpipe;
  
  zok = socketpair(PF_UNIX, SOCK_STREAM, 0, socket_vector);
  if(zok < 0) {
    fprintf(stderr, "socketpair failed: %s\n", strerror(errno));
    return -1;
  }

  /* block sigpipe, as it can fail because the other side is toast. */
  old_sigpipe = signal(SIGPIPE, SIG_IGN);
  if( sfd(connected_sock, socket_vector[1], 8675309) < 0 ) {
    fprintf(stderr, "sendmsg failed: %s\n", strerror(errno));
    close(socket_vector[1]);
    close(socket_vector[0]);
    return -1;
  }
  signal(SIGPIPE, old_sigpipe);

  /* we accepted, so we're connected jus like a connect() ed
     socket would be */
  fish_readhook(socket_vector[0], amphibian_app_write);

  /* set the new socket nonblocking -- prevents us from getting
     locked up, but can lose data. */
  fcntl(socket_vector[0], F_SETFL, O_NONBLOCK);
  
  return(socket_vector[0]);
}

/* this used to have more... */
static void decode_write_errno(int e) {
  if(e == EPIPE) {
    fprintf(stderr, "application terminated unexpectedly\n");
  }
}

/* there is data in the buffer.  the transport protocol has perhaps
   received more of it, or is blocked on our reading it. */
/* TODO: like interrupt-based execution, this is elegant when buffering 
   is low, but not particularly nice when things are blocked */
int amph_receive_data( struct transport_app_state *app,
                       const char *received_data,
                       int valid_bytes) {
  int written = 0;
  if(valid_bytes > 0) {
    /* block sigpipe -- if the app has crashed, we don't
       want to take the signal and croak too */
    siggy_ztardust old_sigpipe = signal(SIGPIPE, SIG_IGN);
    if((written = write(app->fd, received_data, valid_bytes)) < 0) {
      decode_write_errno(errno);
      // TODO: proto->abort_connection(app);
      // TODO: consider RST flag, meanwhile pray it works
      // out instead, amph_close(ts->app.fd);
      if(errno == EWOULDBLOCK) written = 0;
    }
    signal(SIGPIPE, old_sigpipe);
  }
  return written;
}

/* when the other side sends a fin, we will get no more data
   for our app.  calling shutdown causes the app's call to
   read() to unblock, without causing any problems if the
   connection is supposed to remain half-open */
void amph_received_fin( struct transport_app_state *app) {
  shutdown(app->fd, SHUT_WR);
}

/* called on idle_timeout or on connect() failure. */
void amph_destroy ( struct transport_app_state *app ) {
  if(app->ready_for_data == amph_connection_established) {
    /* we were waiting for a connection to be completed. */
    int len = strlen(CONNECT_FAILED);
    int written;
    if((written = write(app->fd, CONNECT_FAILED, len)) < len) {
      if(written < 0) { decode_write_errno(errno); }
      close(app->fd);
      fish_remove_readhook(app->fd);
      set_transport_app_state_for(app->fd, NULL);
      app->fd = -1;
    }
  }
  if(app->fd != -1) {
    close(app->fd);
    fish_remove_readhook(app->fd);
    set_transport_app_state_for(app->fd, NULL);
    app->fd = -1;
  }
}

/* gets called when the transport protocol has received a
   SYN on a port we're interested in, we can pass up the
   socket to the amphibian application because the fishnet
   handshake is simple. */
void accept_callback(struct transport_app_state *app, 
                     int remote_addr,
                     int remote_port,
                     void *user_data) {
  int accepting_socket = (int)user_data;
  struct sockaddr_fish sf;
  int written;

  /* send a new socket to the app, store our end here. */
  app->fd           = amphibian_accept_for_app(accepting_socket);
  app->ready_for_data  = NULL;
  app->received_fin = amph_received_fin;
  app->receive_data = amph_receive_data;
  app->destroy      = amph_destroy;

  /* send the remote address up */
  sf.sfish_family = AF_FISH;
  sf.sfish_address = remote_addr;
  sf.sfish_port = remote_port;
  written = write(app->fd, &sf, sizeof(struct sockaddr_fish));
  if(written < 0) { decode_write_errno(errno); }

  /* store a binding so we can tell which transport_app_state structure
     is used for this application socket */
  set_transport_app_state_for(app->fd, app);

  /* there may have been data in the syn packet, but it's up
     to the transport protocol to deliver it. */
  /* amph_receive_data(app, p->payload, 
     packet_length - sizeof(struct header) - 
     sizeof(struct transport_header)); */
}

/* called when there's something ready on a newly accepted
   amphibian socket: an amphibian application has called 
   bind() or connect(). */
static void amphibian_read_header(int connected_sock) {
  struct connection_request req;
  assert(proto != NULL);

  if(read(connected_sock,  &req, sizeof(req)) < (int)sizeof(req)) {
    fprintf(stderr, "read essentially failed: %s\n", strerror(errno));
    fish_remove_readhook(connected_sock);
    close(connected_sock);
  }
  fish_remove_readhook(connected_sock);

  if(req.connect_not_bind) {
    /* it called connect() */
    struct transport_app_state *app = proto->connect(req.dst.sfish_address, 
                                                     req.dst.sfish_port);
    set_transport_app_state_for(connected_sock, app);

    app->fd   = connected_sock;
    /* when we receive the ack, the connection is established, and
       we can respond with success */
    app->ready_for_data  = amph_connection_established;
    app->received_fin  = amph_received_fin;
    app->receive_data = amph_receive_data;
    /* if it's destroyed, we'll have to say it didn't work */
    app->destroy      = amph_destroy;

    /* when the transport protocol gets an ack, it will 
       call connection_established */
  } else {
    /* app called bind() */
    proto->bind(req.dst.sfish_port, accept_callback, (void *) connected_sock);
    /* when the transport protocol gets a syn, it will call
       accept_callback */
  }
}

/* called when there's something ready on the accepting
   amphibian socket -- a new amphibian application has 
   called socket() */
static void amphibian_accept(int accepting_socket) {
  struct sockaddr_un saun_remote;
  socklen_t slen = sizeof(struct sockaddr_un);
  int connected_sock = accept(accepting_socket, 
                              (struct sockaddr *)&saun_remote, &slen);
  if(connected_sock < 0) {
    fprintf(stderr, "accept failed: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }

  /* set the new socket nonblocking -- prevents us from getting
     locked up, but can lose data. */
  fcntl(connected_sock, F_SETFL, O_NONBLOCK);

  /* now we want to get an informational header from the
     socket to pass as a virtual connect() call to the fish
     transport layers */
  fish_readhook(connected_sock, amphibian_read_header);
}

static void sigpipe_abort(int signum __attribute__((unused))) {
  abort();
}

/* create the unix socket, listen on it, and register our 
   interest in it with a readhook */
int amphibian_init(struct amphibian_protocol *protocol, int fish_address) {
  struct sockaddr_un saun;
  int zok;
  int accepting_socket = socket(PF_UNIX, SOCK_STREAM, 0);

  proto = protocol;

  if(accepting_socket < 0) { 
    fprintf(stderr, "socket failed: %s\n", strerror(errno)); 
    return FALSE; 
  }
  memset(&saun, 0, sizeof(struct sockaddr_un));
  saun.sun_family = AF_UNIX;
  saun.sun_path[0]='\0';
  sprintf(saun.sun_path+1, "fish-%s-%d", getenv("USER"), fish_address);
  zok = bind(accepting_socket, (struct sockaddr *)&saun, sizeof(struct sockaddr_un));
  if(zok < 0) { 
    fprintf(stderr, "bind failed: %s\n", strerror(errno)); 
    return FALSE; 
  }
  zok = listen(accepting_socket, 5);
  if(zok < 0) { 
    fprintf(stderr, "listen failed: %s\n", strerror(errno)); 
    return FALSE; 
  }
  
  fish_readhook(accepting_socket, amphibian_accept);

  signal(SIGPIPE, sigpipe_abort);
  return TRUE;
}


/* vim:set ts=2: */
/*
 * Local Variables:
 * tab-width: 2
 * c-indent-level: 2
 * c-basic-offset: 2
 * End:
 */
