/* fishguts.h - Fishnet internals*/

#ifndef __FISHGUTS_H__
#define __FISHGUTS_H__

/* headers to handle structures listed here */
#include <stdio.h>
#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <sys/timeb.h>
#define bzero(a,len) memset(a,0,len)
#include <io.h>
#define STDIN_FILENO 0
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <unistd.h>
#endif
#include <sys/types.h>

#ifdef __LCLINT__
/*@-incondefs@*/
const char /*@dependent@*/*inet_ntoa(struct in_addr);
int accept(int, /*@out@*/struct sockaddr *, socklen_t *);
/*@+incondefs@*/
#endif

#include "fish.h"

#define FISHGUTS_VERSION_MAJOR (0x02)
#define FISHGUTS_VERSION_MINOR (0x00)

#define DEFAULT_PORT 7777
#define DEFAULT_HOST 0
#define MAX_NEIGHBORS 12
#define FISHBITS_BCST_PORT (8888) // Port for broadcast packets

/*
 * The version header is used to ensure correct interaction between
 * different fishguts versions. Packets with the same major as the
 * fishguts code can be processed; packets with a different major 
 * are dropped. The minor field represents small tweaks to the 
 * interface that still permit interoperability.
 */
struct version {
  unsigned short v_major;  // Packets with the same major are interoperable
  unsigned short v_minor;  // Minor represents small tweaks to the interface
};



/* All user-level packets that are transmitted and received by 
 * fishguts are prepended with the following structure. The header
 * consists of an fishguts version, the intermediate fishnet source
 * address and the intermediate fishnet destination address. The
 * intermediate fishnet addresses are provided to build up a 
 * of (IP addr -> fishnet address) entries based upon packets seen.
 */
struct fishguts_header {
  struct version ver;  // fishguts protocol version number
  int fish_src;		// intermediate source address
  int fish_dst;		// intermediate destination address
};


/*
 * Packet that is actually transmitted between all fishguts nodes.
 */
struct fishguts_packet {
  // Fishguts header
  struct fishguts_header fg_hdr;

  // Packet generated by the user
  struct packet pkt;
};

#define MAX_FISHGUTS_PKT_SIZE (sizeof(struct fishguts_packet))

/*
 * Structure used by fishguts to build up a list of (IP addr -> fishnet addr)
 * entries. This list is consulted whenever a packet is sent to find the
 * IP address associated with a fishhead node.
 *
 * Note that this is currently only used in adhoc mode. As a result, there
 * is no need to capture the port info since all communicating fish nodes
 * must be using the same UDP broadcast port.
 */
struct mapping_entry {
  int fish_addr;
  struct in_addr node_ip_address;
  struct mapping_entry *next;
};



/* 
 * Internal structure used by fishguts to keep track of neighbors that are
 * provided by the fishhead.
 */
struct neighbor {
  int address;                    /* chosen address, HOST order */
  struct in_addr node_ip_address; /* ip address NETWORK order */
  unsigned short node_udp_port;   /* udp address NETWORK order */
  unsigned short lossrate_x_1000; /* loss rate across this link */
};

struct joinrequest {
  int node_address;               /* chosen address, HOST order */
  unsigned short node_udp_port;   /* udp address NETWORK order */
  short reserved;                 /* 32-bit alignment padding */
  char username[8];               /* for ownership reporting */
};

#define FISH_MSG_TYPE_ERROR        0
#define FISH_MSG_TYPE_NEIGHBORLIST 1
#define FISH_MSG_TYPE_JOINREQUEST  2
#define FISH_MSG_TYPE_MAX          3 /* upper bound */

struct fish_message {
  unsigned short length;    /* HOST order (may change later) */
  unsigned short type;
  union {
    /* zero size arrays terminating a message can be
       referenced with a longer array index */
    char errormessage[0]; /* null terminated, 0 included in
                             length */
    struct neighbor neighbors[0]; /* terminated by message
                                     length */
    struct joinrequest joinreq; 
  } body;
  /* put nothing here  (neighbors will be out-of-bounds) */
};

#define MIN_MESSAGE_SIZE 4

/* common routines */
/* the following is gcc specific */
#ifndef WIN32
#define DEBUG_PRINTF(fmt,args...) fprintf(stderr, __FILE__ ":%d" fmt, \
                                                  __LINE__ , ## args )
#else
#define DEBUG_PRINTF(x) do {} while(0)
#endif

/* in fishtimer.c */
void fish_executepending(void);
int fish_time_to_next_event(struct timeval *delta_t);

/* in fishdebug.c */
extern void printmessage(const struct fish_message *pfm);
extern void printneighbors(const struct neighbor *ngh);

void errorabort(const char* msg);
#endif
