/**
 * Hello world example
 * CSE461
 * Autumn 2002
 **/

/*
 * In a shell, type "./fishhead -p <port number>" e.g. "./fishhead -p 7777".
 * <port number> should be between 1024 and 32K. Typing only "fishhead"
 * defaults the <port number> to 7777. 
 * Start up several shells, and in each shell, start a fish node with
 * "./hello <machine name>:<port number> <address of node>" 
 * e.g. "./hello ceylon.cs.washington.edu:7777 4".
 * <machine name> is the name of the machine where fishhead is started, and
 * <address of node> is an integer identifying the node.
 * Next, enter "hi <address of node>" to send a greeting to the node of choice.
 *
 * Note that this program only sends messages to neighboring nodes
 * (within radio distance if in wireless or connected in a wired
 * topology). The first assignment will provide the functionality for
 * sending to non-neighboring nodes.  
 */



#include "fish.h"
#include <string.h>

/* A unique packet identifier is added to each packet. */
static int packetSeqNum = 0;


/*
 * Handle a packet that is meant for me. Print out the string
 * contained within the packet.
 */
static void 
handle_packet(int from, struct packet* p_pkt, int length __attribute__((unused)) ) 
{
  p_pkt->data[PACKET_DATA_SIZE-1] = '\0';
  printf("Node %d says: %s\n", from, p_pkt->data);
}




/*
 * Process the entered text string. If the string is "hi <dest>," 
 * send a packet to <dest> containing the data string "hi :)".
 */
static int
hello_console_command(char *text)
{  
  int dst;
  const char* msg = "hi :)";
  struct packet pkt;

  // Scan through the string to determine if it matches the specified
  // format. If the string contains "hi <dest>," capture the value of
  // <dest>.
  if (1 != sscanf(text, "hi %i[^\n]", &dst)) {
    return FALSE;
  }

  // Create a packet and set all of the control information.
  pkt.hdr.src = fish_getaddress();
  pkt.hdr.dst = dst;
  pkt.hdr.ttl = MAX_TTL;
  pkt.hdr.protocol = FISH_PROTOCOL_PACKET;
  pkt.hdr.pktid = packetSeqNum++;
  pkt.hdr.cksum = 0;

  // Copy the message into the packet; making sure the data array ends
  // with a '\0.'
  strncpy(pkt.data, msg, PACKET_DATA_SIZE-1);
  pkt.data[PACKET_DATA_SIZE-1] = '\0';

  // Send the newly created packet to the specified destination
  fish_send(dst, &pkt, PACKET_HEADER_SIZE + strlen(pkt.data) + 1);

  return TRUE;
}




/* 
 * Handle keyboard input. When a command is entered on the console,
 * it is parsed here and an appropriate action is taken.
 *
 * This function is registered as a callback routine. When a packet
 * arrives, this function is invoked by fish_main() with the entered
 * text.
 */
static void 
on_keyboard(char* text) 
{
  if (hello_console_command(text)) {
    // This was a hello command.
  } else {
    // Unknown command; print error message
    printf("Commands:\n  "
           "hi <dest>\n");
  }
}




/*
 * Print the packet that was received at the provided debug level. This could
 * be very useful for debugging invalid packets that are received.
 */
static void
print_packet(int debuglevel, struct packet *p_pkt)
{
  // Make sure that the data section actually ends with a '\0'
  p_pkt->data[PACKET_DATA_SIZE-1] = '\0';    

  fish_debug(debuglevel,
	     "\tsrc: %d dst: %d ttl: %d proto: %d pktid: %d cksum: %d msg: %s\n",
	     p_pkt->hdr.src, p_pkt->hdr.dst,
	     p_pkt->hdr.ttl, p_pkt->hdr.protocol,
	     p_pkt->hdr.pktid, p_pkt->hdr.cksum,
	     p_pkt->data);
}




/* 
 * Receive a frame and process it.
 *
 * This function is registered as a callback routine. When a packet
 * arrives, this function is invoked by fish_main() with the values
 * corresponding to the packet.
 */
static void 
on_receive(int from, void* frame, int length) 
{
  struct packet *p_pkt = (struct packet *)frame;

  // Checking the protocol field helps to protect us against naked frames
  // or other non-packet data.
  if (p_pkt->hdr.protocol < FISH_PROTOCOL_ERROR || 
      p_pkt->hdr.protocol >= FISH_PROTOCOL_MAX) {
    fish_debug(FISH_DEBUG_NONE,
	       "received frame with invalid protocol; discarding\n");
    return;
  }

  // We should only process packets of type FISH_PROTOCOL_PACKET
  if( p_pkt->hdr.protocol != FISH_PROTOCOL_PACKET ) {
    fish_debug(FISH_DEBUG_NONE,
	       "Received a packet of unsupported protocol: %d.\n"
	       "It was dropped on the floor.\n",
	       p_pkt->hdr.protocol);
    print_packet(FISH_DEBUG_NONE, p_pkt);
    return;
  }

  if( (p_pkt->hdr.dst == fish_getaddress()) ||
      (p_pkt->hdr.dst == ALL_NEIGHBORS) ) {
    // If the packet destination is our address or ALL_NEIGHBORS, process it.
    handle_packet(from, frame, length);
  } else {
    // Packet not destined for us. Drop it on the floor.
    fish_debug(FISH_DEBUG_NONE,
	       "Received a packet with a different node's destination address: %d.\n"
	       "It was dropped on the floor.\n",
	       p_pkt->hdr.dst);
    print_packet(FISH_DEBUG_NONE, p_pkt);
  }
}



/*
 * If the program is invoked with incorrect command line arguments,
 * print out the correct usage.
 */
static void 
print_usage(void) 
{
  printf("Usage: ./hello headhost:headport address\n");
  printf("\theadhost:headport -- specifies the Fishnet node to connect to.\n");
  printf("\taddress -- indicates the Fishnet address of this node.\n");
}




/*
 * The main routine registers callbacks for keyboard input and received
 * packets before joining a fishhead network.
 */
int 
main(int argc, char** argv) 
{
  // Make sure that the program is executed with the correct command line
  // arguments.
  if (argc != 3) {
    print_usage();
    exit(-1);
  }

  // Register a callback routine for keyboard input.
  fish_keybhook(on_keyboard);

  // Register a callback routine for all incoming network packets.
  fish_recvhook(on_receive);

  // Join the Fishhead network.
  fish_joinnetwork(argv[1], atoi(argv[2]));

  // Start the main Fishnet processing loop. This loop is responsible for 
  // detecting incoming packets and keyboard input and invoking the
  // appropriate callback routines.
  fish_main();
  return 0;
}
